import numpy as np
from collections import namedtuple
from hazma.parameters import (
    charged_B_mass as mB,
    charged_kaon_mass as mk,
    charged_pion_mass as mpi,
    muon_mass as mmu,
    B_width,
    k_width,
    kl_width,
)


# Conversion factors
cm_to_inv_MeV = 5.08e10  # MeV^-1 cm^-1

# Convert <sigma v> from MeV^2
sv_inv_MeV_to_cm3_per_s = 1.0 / cm_to_inv_MeV ** 2 * 3e10  # cm^3/s * MeV^2

# Transverse extents of detectors in cm
# arxiv:0709
r_det_E949 = 145.0
# "Muon identification in the Belle experiment at KEKB", Abashian et al, 2002
r_det_belle = 331.0
# "The BABAR Detector", arxiv:0105044, figure 73, page 76
r_det_babar = 320.0

# Vertex resolutions of detectors in cm
# arXiv:hep-ex/0308044
r_vert_belle = 0.5
# arXiv:1405.7808: 13 microns transverse, 71 longitudinal, but 1310.6752,
# 1412.5174 find 5mm! Also need to use boost factor of 20...
r_vert_lhcb = 0.5
# Source: Jason Evans...
r_vert_NA482 = 3.0
# KTeV TDR, page 137, ECAL. TODO: what about 0.01, KTeV TDR, page 119, drift
# chambers???
r_vert_ktev = 0.1

"""Container for meson decay measurements.

Attributes
----------
width_bound : float
    Maximum contribution permitted to the width for this process.
r_max : float
    Mediator decays further than this distance (in cm) from the experiment's
    beampipe are treated as invisible.
s_bounds : list of (float, float)
    Pairs specifying the values of (p_mes - p_prods)^2 measurable by the
    experiment. This is necessary for handling experiments' kinematic cuts.
"""
RareDecayObs = namedtuple("RareDecayObs", ["width_bound", "r_max", "s_bounds"])

# Measurements and predictions for rare decay branching fractions
# B^+ -> K^+ invis, arxiv:1303.7465
br_B_k_invis_SM = (4.5 - 0.7) * 1.0e-6
# BABAR measurement, arxiv:1303.7465
br_B_k_invis_babar = 1.6e-5
B_k_invis_obs = RareDecayObs(
    width_bound=(br_B_k_invis_babar - br_B_k_invis_SM) * B_width,
    r_max=r_det_babar,
    s_bounds=[[0.0, 0.3 * mB ** 2]],
)

# K^+ -> pi^+ invis, arxiv:0709.1000
br_k_pi_invis_SM = (0.85 - 0.07) * 1.0e-10
# E949 measurement, arxiv:0709.1000
br_k_pi_invis_E949 = 1.73e-10 + 1.15e-10
# Bounds on pion momentum must be converted to bounds on s
k_pi_invis_obs = RareDecayObs(
    width_bound=(br_k_pi_invis_E949 - br_k_pi_invis_SM) * k_width,
    r_max=r_det_E949,
    s_bounds=[
        [
            max((mk ** 2 + mpi ** 2 - 2 * mk * np.sqrt(mpi ** 2 + p ** 2)), 0.0)
            for p in p_pis
        ]
        for p_pis in [[229.0, 211.0], [195.0, 140.0]]
    ],
)

# B^+ -> K^+ mu mu, arXiv:hep-ph/0112300
br_B_k_mu_mu_SM = (3.5 - 1.2) * 1.0e-7
# Belle measurement, arxiv:0904.0770
br_B_k_mu_mu_belle = (4.8 + 0.6) * 1.0e-7
B_k_mu_mu_obs = RareDecayObs(
    width_bound=(br_B_k_mu_mu_belle - br_B_k_mu_mu_SM) * B_width,
    r_max=r_vert_belle,
    s_bounds=[[4 * mmu ** 2, 8.68e6], [10.09e6, 12.86e6], [14.18e6, (mB - mk) ** 2]],
)

# B^+ -> K^+ e^+ e^-, arXiv:hep-ph/0112300
br_B_k_e_e_SM = br_B_k_mu_mu_SM
# Belle measurement, arxiv:0904.0770
br_B_k_e_e_belle = (4.8 + 0.6) * 1.0e-7 / 1.03
B_k_e_e_obs = RareDecayObs(
    width_bound=(br_B_k_e_e_belle - br_B_k_e_e_SM) * B_width,
    r_max=r_vert_belle,
    s_bounds=[[8.11e6, 10.03e6], [12.15e6, 14.11e6]],
)

# K_L -> pi^0 mu^+ mu^-, arxiv:hep-ph/0404127
br_kl_pi0_mu_mu_SM = (1.5 - 0.3) * 1.0e-11
# KTeV upper bound, arxiv:hep-ex/0001006
br_kl_pi0_mu_mu_ktev = 3.8e-10
kl_pi0_mu_mu_obs = RareDecayObs(
    width_bound=(br_kl_pi0_mu_mu_ktev - br_kl_pi0_mu_mu_SM) * kl_width,
    r_max=r_vert_ktev,
    s_bounds=[[0.0, 350.0 ** 2]],
)

# K_L -> pi^0 e^+ e^-, arxiv:hep-ph/0308008
br_kl_pi0_e_e_SM = (3.2 - 0.8) * 1.0e-11
# KTeV upper bound, arxiv:hep-ex/0309072
br_kl_pi0_e_e_ktev = 2.8e-10
kl_pi0_e_e_obs = RareDecayObs(
    width_bound=(br_kl_pi0_e_e_ktev - br_kl_pi0_e_e_SM) * kl_width,
    r_max=r_vert_ktev,
    s_bounds=[[140.0 ** 2, 362.7 ** 2]],
)

"""Container for information about beam dump experiments.

Attributes
----------
br_pN_k : float
    Inclusive branching fraction for pN -> K, where K is a charged or neutral
    kaon.
br_pN_B : float
    Inclusive branching fraction for pN -> K, where B is a charged or neutral
    B-meson.
mediator_energy : float
    Energy of particle produced in meson decays (MeV).
dist : float
    Distance to the detector (m).
len : float
    Length of the detector (m).
visible_fss : list of strings
    List of mediator decay final states that can be detected.
n_pot : float
    Number of protons on target.
n_dec : float
    Upper bound on the number of mediator decays.
"""
BeamDumpParams = namedtuple(
    "BeamDumpParams",
    [
        "br_pN_k",
        "br_pN_B",
        "mediator_energy",
        "dist",
        "length",
        "visible_fss",
        "n_pot",
        "n_dec",
    ],
)

# CHARM experiment
# Dolan et al (arXiv:1412.5174) use br_pN_k=3/7, which is a ~10% difference.
bd_charm = BeamDumpParams(
    br_pN_k=10.0 / 40.0,
    br_pN_B=3.6e-6 / 40.0,
    mediator_energy=10.0e3,
    dist=480.0,
    length=35.0,
    visible_fss=["g g", "e e", "mu mu"],
    n_pot=2.9e17,
    n_dec=2.3,
)

# SHiP experiment
bd_ship = BeamDumpParams(
    br_pN_k=10.0 / 40.0,
    br_pN_B=3.6e-3 / 40.0,
    mediator_energy=25.0e3,
    dist=70.0,
    length=55.0,
    visible_fss=["g g", "e e", "mu mu", "pi pi", "pi0 pi0", "k k", "k0 k0"],
    n_pot=2.0e20,
    n_dec=3.0,
)



from cmath import sqrt, pi, atanh
from hazma.parameters import charged_pion_mass as mpi
from hazma.parameters import neutral_pion_mass as mpi0
from hazma.parameters import muon_mass as mmu
from hazma.parameters import electron_mass as me
from hazma.parameters import fpi, qe
from scipy.integrate import quad


class VectorMediatorCrossSections:
    def sigma_xx_to_v_to_ff(self, e_cm, f):
        """
        Returns the cross section for xbar x to fbar f.

        Parameters
        ----------
        e_cm : float
            Center of mass energy.
        f : float
            Name of final state fermion: "e" or "mu".
        self : object
            Class containing the vector mediator parameters.

        Returns
        -------
        cross_section : float
            Cross section for xbar + x -> v -> fbar + f.
        """
        if f == "e":
            mf = me
            gvll = self.gvee
        elif f == "mu":
            mf = mmu
            gvll = self.gvmumu

        mx = self.mx

        if e_cm >= 2.0 * mf and e_cm >= 2.0 * mx:
            gvxx = self.gvxx
            mv = self.mv
            width_v = self.width_v

            ret_val = (
                              gvll**2
                              * gvxx**2
                              * sqrt(e_cm**2 - 4.0 * mf**2)
                              * (e_cm**2 + 2.0 * mf**2)
                              * (e_cm**2 + 2 * mx**2)
                      ) / (
                              12.0
                              * e_cm**2
                              * sqrt(e_cm**2 - 4.0 * mx**2)
                              * pi
                              * (
                                      e_cm**4
                                      - 2.0 * e_cm**2 * mv**2
                                      + mv**4
                                      + mv**2 * width_v**2
                              )
                      )

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def sigma_xx_to_v_to_pipi(self, e_cm):
        """
        Returns the cross section for xbar x to pi+ pi-.

        Parameters
        ----------
        e_cm : float
            Center of mass energy.
        self : object
            Class containing the vector mediator parameters.

        Returns
        -------
        cross_section : float
            Cross section for xbar + x -> v -> f + f.
        """
        mx = self.mx

        if e_cm >= 2.0 * mpi and e_cm >= 2.0 * mx:
            gvuu = self.gvuu
            gvdd = self.gvdd
            gvxx = self.gvxx
            mv = self.mv
            width_v = self.width_v

            ret_val = (
                              (gvdd - gvuu)**2
                              * gvxx**2
                              * (-4.0 * mpi**2 + e_cm**2)**1.5
                              * (2.0 * mx**2 + e_cm**2)
                      ) / (
                              48.0
                              * pi
                              * e_cm**2
                              * sqrt(-4.0 * mx**2 + e_cm**2)
                              * (
                                      mv**4
                                      - 2.0 * mv**2 * e_cm**2
                                      + e_cm**4
                                      + mv**2 * width_v**2
                              )
                      )

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def sigma_xx_to_v_to_pi0g(self, e_cm):
        """
        Returns the cross section for xbar x to pi0 g.

        Parameters
        ----------
        e_cm : float
            Center of mass energy.
        self : object
            Class containing the vector mediator parameters.

        Returns
        -------
        cross_section : float
            Cross section for xbar + x -> v -> pi0 g
        """
        mx = self.mx

        if e_cm >= mpi0 and e_cm >= 2.0 * mx:
            gvuu = self.gvuu
            gvdd = self.gvdd
            gvxx = self.gvxx
            mv = self.mv
            width_v = self.width_v

            ret_val = (
                    3.0
                    * (
                            (gvdd + 2.0 * gvuu)**2.0
                            * gvxx**2
                            * (-mpi0**2 + e_cm**2)**3
                            * (2.0 * mx**2 + e_cm**2)
                            * qe**2
                    )
                    / (
                            13824.0
                            * fpi**2
                            * pi**5
                            * e_cm**3
                            * sqrt(-4.0 * mx**2 + e_cm**2)
                            * (
                                    mv**4
                                    - 2.0 * mv**2 * e_cm**2
                                    + e_cm**4
                                    + mv**2 * width_v**2
                            )
                    )
            )

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def sigma_xx_to_v_to_pi0v(self, e_cm):
        """
        Returns the cross section for xbar x to pi0 v.

        Parameters
        ----------
        e_cm : float
            Center of mass energy.
        self : object
            Class containing the vector mediator parameters.

        Returns
        -------
        cross_section : float
            Cross section for xbar + x -> v -> pi0 v
        """
        mx = self.mx
        mv = self.mv

        if e_cm >= mpi0 + mv and e_cm >= 2.0 * mx:
            gvuu = self.gvuu
            gvdd = self.gvdd
            gvxx = self.gvxx
            width_v = self.width_v

            ret_val = ((gvdd - gvuu)**2 * (gvdd + gvuu)**2 * gvxx**2 * (
                    (-mpi0 - mv + e_cm) *
                    (mpi0 - mv + e_cm) *
                    (-mpi0 + mv + e_cm) *
                    (mpi0 + mv + e_cm))**1.5 * (2 * mx**2 + e_cm**2)) / (
                              1536.0 * fpi**2 * pi**5 * e_cm**3 *
                              sqrt(-4 * mx**2 + e_cm**2)
                              * ((-mv**2 + e_cm**2)**2 + mv**2 * width_v**2)
                      )

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def dsigma_ds_xx_to_v_to_pi0pipi(self, s, e_cm):
        mx = self.mx

        if (e_cm > 2.0 * mpi + mpi0 and e_cm > 2.0 * mx
                and 4.0 * mpi**2 < s < (e_cm - mpi0)**2):
            gvuu = self.gvuu
            gvdd = self.gvdd
            gvxx = self.gvxx
            mv = self.mv
            width_v = self.width_v

            ret_val = (
                    3.0
                    * (
                            (gvdd + gvuu)**2
                            * gvxx**2
                            * sqrt(s * (-4.0 * mpi**2 + s))
                            * sqrt(
                        e_cm**4
                        + (mpi0**2 - s)**2
                        - 2.0 * e_cm**2 * (mpi0**2 + s)
                    )
                            * (
                                    -24.0 * mpi**6 * s
                                    + mpi**4
                                    * (-2.0 * mpi0**4 + 28.0 * mpi0**2 * s + 22.0 * s**2)
                                    + 2.0 * mpi**2 * (mpi0**6 - 4.0 * s**3)
                                    + s
                                    * (
                                            -2.0 * mpi0**6
                                            - 4.0 * mpi0**4 * s
                                            - mpi0**2 * s**2
                                            + s**3
                                    )
                                    + e_cm**4
                                    * (
                                            -2.0 * mpi**4
                                            + 2.0 * mpi**2 * (mpi0**2 - s)
                                            + s * (-2.0 * mpi0**2 + s)
                                    )
                                    + e_cm**2
                                    * (
                                            4.0 * mpi**4 * (mpi0**2 + s)
                                            + s * (4.0 * mpi0**4 + 5.0 * mpi0**2 * s - 2.0 * s**2)
                                            - 4.0
                                            * mpi**2
                                            * (mpi0**4 + 3.0 * mpi0**2 * s - s**2)
                                    )
                            )
                    )
                    / (
                            294912.0
                            * fpi**6
                            * pi**7
                            * sqrt(e_cm**2)
                            * sqrt(-4.0 * mx**2 + e_cm**2)
                            * s**2
                            * (
                                    mv**4
                                    - 2.0 * mv**2 * e_cm**2
                                    + e_cm**4
                                    + mv**2 * width_v**2
                            )
                    )
            )

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def sigma_xx_to_v_to_pi0pipi(self, e_cm):
        if e_cm > 2.0 * mpi + mpi0 and e_cm > 2.0 * self.mx:
            s_min = 4.0 * mpi**2
            s_max = (e_cm - mpi0)**2

            ret_val = quad(
                self.dsigma_ds_xx_to_v_to_pi0pipi, s_min, s_max, args=(e_cm)
            )[0]

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def sigma_xx_to_vv(self, e_cm):
        mx = self.mx
        mv = self.mv

        if e_cm >= 2.0 * mv and e_cm >= 2.0 * mx:
            gvxx = self.gvxx

            ret_val = (
                              gvxx**4
                              * sqrt(-4 * mv**2 + e_cm**2)
                              * (
                                      -2
                                      - (2 * (mv**2 + 2 * mx**2)**2)
                                      / (mv**4 - 4 * mv**2 * mx**2 + mx**2 * e_cm**2)
                                      + (
                                              4
                                              * (
                                                      4 * mv**4
                                                      - 8 * mv**2 * mx**2
                                                      - 8 * mx**4
                                                      + 4 * mx**2 * e_cm**2
                                                      + e_cm**4
                                              )
                                              * atanh(
                                          (
                                                  sqrt(4 * mv**2 - e_cm**2)
                                                  * sqrt(4 * mx**2 - e_cm**2)
                                          )
                                          / (2 * mv**2 - e_cm**2)
                                      )
                                      )
                                      / (
                                              (2 * mv**2 - e_cm**2)
                                              * sqrt(4 * mv**2 - e_cm**2)
                                              * sqrt(4 * mx**2 - e_cm**2)
                                      )
                              )
                      ) / (8.0 * pi * e_cm**2 * sqrt(-4 * mx**2 + e_cm**2))

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def sigma_xx_to_v_to_xx(self, e_cm):
        """
        Returns the DM annihilation cross section into DM.
        """
        rv = self.mv / e_cm
        rx = self.mx / e_cm
        gvxx = self.gvxx
        rwv = self.width_v / e_cm

        if e_cm > 2.0 * self.mx:

            def msqrd(z):
                return (
                               gvxx**4
                               * (
                                       9
                                       - 56 * rx**2
                                       + 2
                                       * (
                                               9 * rv**2 * (-1 + rv**2 + rwv**2)
                                               + 4 * rv**2 * (11 + 2 * rv**2 + 2 * rwv**2) * rx**2
                                               + 8
                                               * (7 + 6 * rv**2 * (-4 + rv**2 + rwv**2))
                                               * rx**4
                                               + 64 * rx**6
                                       )
                                       - 2
                                       * (-1 + 4 * rx**2)
                                       * (
                                               3 * rv**2 * (-3 + 2 * rv**2 + 2 * rwv**2)
                                               + 4
                                               * (3 + 6 * rv**4 + 2 * rv**2 * (-7 + 3 * rwv**2))
                                               * rx**2
                                               - 32 * (-2 + rv**2) * rx**4
                                       )
                                       * z
                                       + 2
                                       * (3 * rv**2 - 4 * rx**2)
                                       * (-1 + 4 * rx**2)**3
                                       * z**3
                                       + (1 - 4 * rx**2)**4 * z**4
                                       + 2
                                       * (
                                               3
                                               + 5 * rv**4
                                               + 12 * rx**2
                                               + 8 * rx**4
                                               + rv**2 * (-3 + 5 * rwv**2 - 20 * rx**2)
                                       )
                                       * (z - 4 * rx**2 * z)**2
                               )
                       ) / (
                               (1 + rv**4 + rv**2 * (-2 + rwv**2))
                               * (
                                       4 * rv**4
                                       + 4 * rv**2 * (rwv**2 + (-1 + 4 * rx**2) * (-1 + z))
                                       + (1 - 4 * rx**2)**2 * (-1 + z)**2
                               )
                       )

            ret_val = quad(msqrd, -1, 1)[0] / (32.0 * pi * e_cm)

            assert ret_val.imag == 0.0
            assert ret_val.real >= 0.0

            return ret_val.real
        else:
            return 0.0

    def annihilation_cross_section_funcs(self):
        return {
            "mu mu": lambda e_cm: self.sigma_xx_to_v_to_ff(e_cm, "mu"),
            "e e": lambda e_cm: self.sigma_xx_to_v_to_ff(e_cm, "e"),
            "pi pi": self.sigma_xx_to_v_to_pipi,
            "pi0 g": self.sigma_xx_to_v_to_pi0g,
            "pi0 v": self.sigma_xx_to_v_to_pi0v,
            "v v": self.sigma_xx_to_vv,
        }



from collections import namedtuple
import numpy as np
from pkg_resources import resource_filename
from hazma.background_model import BackgroundModel
from hazma.flux_measurement import FluxMeasurement
from hazma.parameters import load_interp

"""
Parameters relevant to computing constraints from gamma ray experiments.
"""

# Get paths to files inside the module
gr_data_dir = "gamma_ray_data/"
# Gamma-ray observations
egret_obs_rf = resource_filename(__name__, gr_data_dir + "egret_obs.dat")
comptel_obs_rf = resource_filename(__name__, gr_data_dir + "comptel_obs.dat")
fermi_obs_rf = resource_filename(__name__, gr_data_dir + "fermi_obs.dat")
# Effective areas
A_eff_e_astrogam_rf = resource_filename(
    __name__, gr_data_dir + "e-astrogam_effective_area.dat"
)
A_eff_comptel_rf = resource_filename(
    __name__, gr_data_dir + "comptel_effective_area.dat"
)
A_eff_egret_rf = resource_filename(__name__, gr_data_dir + "egret_effective_area.dat")
A_eff_fermi_rf = resource_filename(__name__, gr_data_dir + "fermi_effective_area.dat")

e_astrogam_energy_res_rf = resource_filename(
    __name__, gr_data_dir + ("e-astrogam_energy" "_resolution.dat")
)
# Complex background model
gc_bg_model_rf = resource_filename(__name__, gr_data_dir + "gc_bg_model.dat")


def solid_angle(l_max, b_min, b_max):
    """
    Returns solid angle subtended for a rectangular target region centered on
    the galactic center.

    Parameters
    ----------
    l_max : float
        Maximum value of galactic longitude in deg. Note that :math:`l` must
        lie in the interval :math:`[-180, 180]`.
    b_min, b_max : float, float
        Minimum and maximum values for :math:`b` in deg. Note that :math:`b`
        must lie in the interval :math:`[-90, 90]`, with the equator at
        :math:`b = 0`.

    Returns
    -------
    Omega : float
        Solid angle subtended by the region in sr.
    """
    deg_to_rad = np.pi / 180.0
    return (
        4.0
        * l_max
        * deg_to_rad
        * (np.sin(b_max * deg_to_rad) - np.sin(b_min * deg_to_rad))
    )


class TargetParams:
    """
    Container for information about a target region.

    Currently implemented for the Draco dwarf galaxy and :math:`10^\circ \times
    10^\circ` region around the galactic center, which can be imported using::

        from hazma.gamma_ray_parameters import draco_params, gc_target

    Parameters
    ----------
    J : float
        J-factor in MeV^2 cm^-5
    dOmega : float
        Angular size in sr
    """

    def __init__(self, J, dOmega):
        self.J = J
        self.dOmega = dOmega


# Dwarf with high J factor
draco_params = TargetParams(6.94e27, 1.62e-3)

# This is the background model from arXiv:1504.04024, eq. 14. It was derived
# by performing a simple power law fit to COMPTEL data from 0.8 - 30 MeV and
# EGRET data from 30 MeV - 10 GeV. We take the lower range of validity to be
# the lowest energy for which e-ASTROGAM has nonzero effective area.
default_bg_model = BackgroundModel([0.3, 10.0e3], lambda e: 2.74e-3 / e ** 2)

# This is the more complex background model from arXiv:1703.02546. Note that it
# is only applicable to the inner 10deg x 10deg region of the Milky Way.
gc_bg_flux_fn = load_interp(gc_bg_model_rf, bounds_error=True, fill_value=np.nan)
gc_bg_e_range = gc_bg_flux_fn.x[[0, -1]]
gc_bg_model = BackgroundModel(gc_bg_e_range, gc_bg_flux_fn)
gc_target = TargetParams(1.795e29, solid_angle(10.0, 0.0, 10.0))

# # # Effective areas, cm^2
A_eff_e_astrogam = load_interp(
    A_eff_e_astrogam_rf
)  #: e-ASTROGAM effective area function
A_eff_fermi = load_interp(A_eff_fermi_rf)  #: Fermi-LAT effective area function
A_eff_comptel = load_interp(A_eff_comptel_rf)  #: COMPTEL effective area function
A_eff_egret = load_interp(A_eff_egret_rf)  #: EGRET effective area function


def energy_res_comptel(e):
    """COMPTEL energy resolution :math:`\Delta E / E`.

    Taken from `ch. II, page 11
    <http://wwwgro.unh.edu/users/ckappada/thesis_stuff/thesis.html>`_.
    """
    return 0.05


def energy_res_egret(e):
    """EGRET's energy resolution :math:`\Delta E / E`.

    This is the most optimistic value, taken from
    `sec. 4.3.3 <http://adsabs.harvard.edu/doi/10.1086/191793>`_.
    """
    return 0.18


def energy_res_fermi(e):
    """Fermi-LAT's energy resolution :math:`\Delta E / E`.

    This is the average of the most optimistic normal and 60deg off-axis values
    from `fig. 18 <https://arxiv.org/abs/0902.1089>`_.
    """
    return 0.075


#: e-ASTROGAM energy resolution function. From table 1 of the `e-ASTROGAM
#: whitebook <https://arxiv.org/abs/1711.01265>`_.
energy_res_e_astrogam = load_interp(e_astrogam_energy_res_rf, fill_value="extrapolate")

# Approximate observing time for e-ASTROGAM in seconds
T_obs_e_astrogam = 365.0 * 24.0 * 60.0 ** 2

# # # Target parameters
comptel_diffuse_target = TargetParams(J=3.725e28, dOmega=solid_angle(60.0, 0.0, 20.0))
egret_diffuse_target = TargetParams(J=3.79e27, dOmega=solid_angle(180.0, 20.0, 60.0))
fermi_diffuse_target = TargetParams(J=4.698e27, dOmega=solid_angle(180.0, 8.0, 90.0))
#: COMPTEL diffuse gamma-ray flux measurements
comptel_diffuse = FluxMeasurement(
    comptel_obs_rf, energy_res_comptel, comptel_diffuse_target
)
#: EGRET diffuse gamma-ray flux measurements
egret_diffuse = FluxMeasurement(egret_obs_rf, energy_res_egret, egret_diffuse_target)
#: Fermi diffuse gamma-ray flux measurements
fermi_diffuse = FluxMeasurement(fermi_obs_rf, energy_res_fermi, fermi_diffuse_target)



from hazma.constraint_parameters import sv_inv_MeV_to_cm3_per_s
from skimage import measure

import numpy as np


class TheoryConstrain:
    def custom_constrain(self, param_grid, ls_or_img="image"):
        """Computes constraints over grid of parameter values.

        Parameters
        ----------
        param_grid : 2D array of parameters
            Parameter values at which to compute constraints.
        ls_or_img : "image" or "ls"
            Controls whether this function returns level sets or images.

        Returns
        -------
        constrs : dict
            A dictionary containing the constraints on the theory in the (p1,
            p2) plane.

            If ls_or_img is "ls", the values are level sets. A level set is a
            list of curves, where each curve is a list of values of (p1, p2)
            defining the parameter values that saturate the constraint. If
            ls_or_img is "image", each value is a 2D numpy.array I(x,y) such
            that I_ij > 0 when (p1_vals[i], p2_vals[j]) is not excluded by the
            corresponding constraint and I_ij < 0 if (p1_vals[i], p2_vals[j])
            is excluded by the constraint.
        """
        n_rows, n_cols = param_grid.shape
        constraints = self.constraints()

        # Store the constraint images. Note that p1 and p2 must be swapped
        # so we can use Cartesian rather than matrix indexing.
        imgs = {cn: np.zeros([n_rows, n_rows]) for cn in constraints.keys()}

        # Loop over the parameter grid
        for i in range(n_rows):
            for j in range(n_cols):
                # Set this theory's parameters to the values at this point
                self.__dict__.update(param_grid[i, j].__dict__)

                # Compute all constraints at this point in parameter space
                for cn, fn in constraints.items():
                    imgs[cn][i, j] = fn()

        if ls_or_img == "image":
            return imgs
        elif ls_or_img == "ls":
            return {cn: self._img_to_ls(img) for cn, img in imgs}

    def constrain(self, p1, p1_vals, p2, p2_vals, ls_or_img="image"):
        """Computes constraints over 2D slice of parameter space.

        Parameters
        ----------
        p1 : string
            Name of a parameter to constraint.
        p1_vals : np.array
            Values of p1 at which to compute constraints. Must be sorted.
        p2 : string
            Name of the other parameter to constraint. Must be different than
            p1.
        p2_vals : np.array
            Values of p2 at which to compute constraints. Must be sorted.
        ls_or_img : "image" or "ls"
            Controls whether this function returns level sets or images.

        Returns
        -------
        constrs : dict
            A dictionary containing the constraints on the theory in the (p1,
            p2) plane.

            If ls_or_img is "ls", the values are level sets. A level set is a
            list of curves, where each curve is a list of values of (p1, p2)
            defining the parameter values that saturate the constraint. If
            ls_or_img is "image", each value is a 2D numpy.array I(x,y) such
            that I_ij > 0 when (p1_vals[i], p2_vals[j]) is not excluded by the
            corresponding constraint and I_ij < 0 if (p1_vals[i], p2_vals[j])
            is excluded by the constraint.
        """
        if p1 == p2:
            raise ValueError(
                "Parameters being constrained must not be the "
                "same. Both are %s." % p1
            )

        n_p1s, n_p2s = len(p1_vals), len(p2_vals)
        constraints = self.constraints()

        # Store the constraint images. Note that p1 and p2 must be swapped
        # so we can use Cartesian rather than matrix indexing.
        imgs = {cn: np.zeros([n_p2s, n_p1s]) for cn in constraints.keys()}

        # Loop over the parameter values
        for idx_p1, p1_val in np.ndenumerate(p1_vals):
            for idx_p2, p2_val in np.ndenumerate(p2_vals):
                setattr(self, p1, p1_val)
                setattr(self, p2, p2_val)

                # Compute all constraints at this point in parameter space
                for cn, fn in constraints.items():
                    imgs[cn][idx_p2[0], idx_p1[0]] = fn()

        if ls_or_img == "image":
            return imgs
        elif ls_or_img == "ls":
            return {cn: self._img_to_ls(img) for cn, img in imgs}

    def constrain_gamma_helper(self, p2, p2_vals, measurement, n_sigma=2):
        """Computes constraints on a parameter from gamma ray experiments.

        Notes
        -----
        * p2 must not depend on mx or the center of mass energy.
        * TODO: make sure this is working correctly, it currently doesn't
        account for gamma ray lines!!!
        """
        vx = 1.0e-3
        e_cm = 2.0 * self.mx * (1 + 0.5 * vx ** 2)  # TODO: should depend on target

        # Factor to convert dN/dE to Phi/<sigma v>
        dm_flux_factor = (
            measurement.target.J
            * measurement.target.dOmega
            / (2.0 * 4.0 * np.pi * self.mx ** 2)
        )

        # Energy range over which to compute convolved spectrum
        e_bin_min, e_bin_max = measurement.bins[0][0], measurement.bins[-1][1]

        def get_bin_fluxes(spec_fn, line_fn):
            """Gets Phi/<sigma v> for a particular channel.
            """
            dnde_det = self.get_detected_spectrum_function(
                e_bin_min, e_bin_max, e_cm, measurement.energy_res, 500
            )
            return np.array(
                [
                    dm_flux_factor * dnde_det.integral(bl, br)
                    for bl, br in measurement.bins
                ]
            )

        # Compute Phi/<sigma v> in each bin for each final state
        fs_bin_fluxes = {
            fs: get_bin_fluxes(spec_fn, lambda e_cm: {})
            for fs, spec_fn in self.spectrum_funcs().items()
            if fs != "total"
        }
        # line_bin_fluxes = {fs: get_bin_fluxes(None, line_fn) for fs, line_fn
        #                    in self.gamma_ray_lines(cme) if fs != "total"}

        def flux_difference(p2, p2_val):
            """Compute difference between Phi_obs+N*sigma - Phi_th."""
            setattr(self, p2, p2_val)

            # Compute cross sections
            css = self.annihilation_cross_sections(e_cm)
            # Get fluxes by multiplying <sigma v>
            bin_fluxes = np.array(
                [
                    bf * css[fs] * vx * sv_inv_MeV_to_cm3_per_s
                    for fs, bf in fs_bin_fluxes.items()
                ]
            )

            return np.min(
                measurement.fluxes
                + n_sigma * measurement.upper_errors
                - bin_fluxes.sum(axis=0)
            )

        return np.array([flux_difference(p2, p2v) for p2v in p2_vals])

    def constrain_gamma(self, p1, p1_vals, p2, p2_vals, measurement, n_sigma=2):
        """Computes constraints from gamma ray experiments in the p1-p2 plane.

        Notes
        -----
        p1 must not depend on mx or the center of mass energy.
        """
        img = np.zeros([len(p2_vals), len(p1_vals)])

        for idx_p1, p1_val in enumerate(p1_vals):
            setattr(self, p1, p1_val)

            # Compute constraint function along the current column
            img[idx_p1, :] = self.constrain_gamma_helper(
                p2, p2_vals, measurement, n_sigma
            )

        return img

    def _img_to_ls(self, p1_vals, p2_vals, img):
        """Finds levels sets for an image.
        """
        contours_raw = measure.find_contours(img, level=0)
        contours = []

        # Convert from indices to values of p1 and p2
        for c in contours_raw:
            p1s = c[:, 1] / len(p1_vals) * (p1_vals[-1] - p1_vals[0]) + p1_vals[0]
            p2s = c[:, 0] / len(p2_vals) * (p2_vals[-1] - p2_vals[0]) + p2_vals[0]
            contours.append(np.array([p1s, p2s]))

        return contours



from setuptools import setup, Extension, find_packages

import numpy as np

VERSION = "1.0.1"


def readme():
    """print long description"""
    with open("READMEPYPI.md") as f:
        return f.read()


decay_dir = "hazma/decay_helper_functions/"
gr_dir = "hazma/gamma_ray_helper_functions/"
ps_dir = "hazma/phase_space_helper_functions/"
ft_dir = "hazma/field_theory_helper_functions/"
pos_dir = "hazma/positron_helper_functions/"
sm_dir = "hazma/scalar_mediator/"
vm_dir = "hazma/vector_mediator/"

decay_pack = "hazma.decay_helper_functions"
ft_pack = "hazma.field_theory_helper_functions"
gr_pack = "hazma.gamma_ray_helper_functions"
theory_pack = "hazma.theory"
ps_pack = "hazma.phase_space_helper_functions"
pos_pack = "hazma.positron_helper_functions"
sm_pack = "hazma.scalar_mediator"
vm_pack = "hazma.vector_mediator"

psm_pack = "hazma.pseudo_scalar_mediator"
avm_pack = "hazma.axial_vector_mediator"
unit_pack = "hazma.unitarization"

packs = [
    "hazma",
    avm_pack,
    decay_pack,
    ft_pack,
    gr_pack,
    theory_pack,
    ps_pack,
    pos_pack,
    psm_pack,
    sm_pack,
    unit_pack,
    vm_pack,
]

extensions = []

# Decay helper functions extensions
extensions += [
    Extension(
        decay_pack + ".decay_charged_pion",
        sources=[decay_dir + "decay_charged_pion.pyx"],
    )
]
extensions += [
    Extension(
        decay_pack + ".decay_neutral_pion",
        sources=[decay_dir + "decay_neutral_pion.pyx"],
    )
]
extensions += [
    Extension(decay_pack + ".decay_muon", sources=[decay_dir + "decay_muon.pyx"])
]
extensions += [
    Extension(
        decay_pack + ".decay_charged_kaon",
        sources=[decay_dir + "decay_charged_kaon.pyx"],
    )
]
extensions += [
    Extension(
        decay_pack + ".decay_long_kaon", sources=[decay_dir + "decay_long_kaon.pyx"]
    )
]
extensions += [
    Extension(
        decay_pack + ".decay_short_kaon", sources=[decay_dir + "decay_short_kaon.pyx"]
    )
]

# Gamma-Ray Helper functions extensions
extensions += [
    Extension(
        gr_pack + ".gamma_ray_generator", sources=[gr_dir + "gamma_ray_generator.pyx"]
    )
]

# Phase space helper functions extensions
extensions += [
    Extension(
        ps_pack + ".generator",
        sources=[ps_dir + "generator.pyx"],
        extra_compile_args=["-g", "-std=c++11"],
        language="c++",
    )
]
extensions += [
    Extension(
        ps_pack + ".histogram",
        sources=[ps_dir + "histogram.pyx"],
        extra_compile_args=["-g", "-std=c++11"],
        language="c++",
    )
]
extensions += [
    Extension(
        ps_pack + ".modifiers",
        sources=[ps_dir + "modifiers.pyx"],
        extra_compile_args=["-g", "-std=c++11"],
        language="c++",
    )
]

# Field Theory helper functions extensions
extensions += [
    Extension(ft_pack + ".common_functions", sources=[ft_dir + "common_functions.pyx"])
]
extensions += [
    Extension(
        ft_pack + ".three_body_phase_space",
        sources=[ft_dir + "three_body_phase_space.pyx"],
    )
]

# Positron Helper functions
extensions += [
    Extension(
        pos_pack + ".positron_charged_pion",
        sources=[pos_dir + "positron_charged_pion.pyx"],
    )
]
extensions += [
    Extension(pos_pack + ".positron_muon", sources=[pos_dir + "positron_muon.pyx"])
]
extensions += [
    Extension(pos_pack + ".positron_decay", sources=[pos_dir + "positron_decay.pyx"])
]

# Scalar mediator
extensions += [
    Extension(
        sm_pack + ".scalar_mediator_decay_spectrum",
        sources=[sm_dir + "scalar_mediator_decay_spectrum.pyx"],
    )
]
extensions += [
    Extension(
        sm_pack + ".scalar_mediator_positron_spec",
        sources=[sm_dir + "scalar_mediator_positron_spec.pyx"],
    )
]
# Vector mediator
extensions += [
    Extension(
        vm_pack + ".vector_mediator_decay_spectrum",
        sources=[vm_dir + "vector_mediator_decay_spectrum.pyx"],
    )
]
extensions += [
    Extension(
        vm_pack + ".vector_mediator_positron_spec",
        sources=[vm_dir + "vector_mediator_positron_spec.pyx"],
    )
]

setup(
    name="hazma",
    version=VERSION,
    author="Logan Morrison and Adam Coogan",
    author_email="loanmorr@ucsc.edu",
    maintainer="Logan Morrison",
    maintainer_email="loanmorr@ucsc.edu",
    url="http://hazma.readthedocs.io",
    description="Python package for computing indirect detection constraints on sub-GeV dark matter.",
    long_description=readme(),
    long_description_content_type="text/markdown",
    keywords="dark-matter mev-dark-matter gamma-ray-spectra",
    packages=find_packages(),
    ext_modules=extensions,
    include_dirs=[
        np.get_include(),
        "hazma/decay_helper_functions",
        "hazma/positron_helper_functions",
    ],
    package_data={
        "hazma/decay_helper_functions": ["*.pxd"],
        "hazma/positron_helper_functions": ["*.pxd"],
    },
    setup_requires=["pytest-runner"],
    install_requires=[
        "pip>=9.0.1",
        "matplotlib>=2.2.3",
        "scipy>=1.0.0",
        "numpy>=1.13.3",
        "cython>=0.27.3",
        "numpydoc>=0.7.0",
        "scikit-image>=0.14",
        "setuptools>=27.3",
        "flake8>=3.5.0",
    ],
    tests_require=["pytest>=3.2.5"],
    zip_safe=False,
    include_package_data=True,
    license="gpl-3.0",
    platforms="MacOS and Linux",
    download_url="https://github.com/LoganAMorrison/Hazma",
    classifiers=["Programming Language :: Python"],
)



from hazma.decay_helper_functions.decay_charged_pion \
    cimport CSpectrum as cp_spec
from hazma.decay_helper_functions.decay_neutral_pion \
    cimport CSpectrum as np_spec
from hazma.decay_helper_functions.decay_neutral_pion cimport CSpectrumPoint
from hazma.decay_helper_functions.decay_muon cimport CSpectrum as mu_spec

import cython
import numpy as np
cimport numpy as np
from scipy.integrate import quad

from libc.math cimport exp, log, M_PI, log10, sqrt, abs, pow

include "../decay_helper_functions/parameters.pxd"

cdef double mmu = MASS_MU
cdef double me = MASS_E
cdef double mpi = MASS_PI
cdef double mpi0 = MASS_PI0
cdef double qe = sqrt(4. * M_PI * ALPHA_EM)
ctypedef np.ndarray ndarray

cdef int n_interp_pts = 500

cdef np.ndarray __e_gams = np.zeros((n_interp_pts,), dtype=np.float64)

cdef np.ndarray __spec_cp = np.zeros((n_interp_pts,), dtype=np.float64)
cdef np.ndarray __spec_mu = np.zeros((n_interp_pts,), dtype=np.float64)

cdef double __set_spectra(double mv):
    global __e_gams
    global __spec_cp
    global __spec_mu

    __e_gams = np.logspace(-1., np.log10(mv / 2.), num=n_interp_pts)

    __spec_cp = cp_spec(__e_gams, mv / 2., "total")
    __spec_mu = mu_spec(__e_gams, mv / 2.)

cdef double __interp_spec(double eng_gam, str mode):
    """
    Intepolation function for the short kaon data.

    Parameters
    ----------
    eng_gam : double
        Energy of the photon.
    mode : str {"total"}
        String specifying which decay mode to use.
    """
    if mode == "cp":
        if eng_gam < 10**-1:
            return __spec_cp[0] * __e_gams[0] / eng_gam
        return np.interp(eng_gam, __e_gams, __spec_cp)
    if mode == "mu":
        if eng_gam < 10**-1:
            return __spec_mu[0] * __e_gams[0] / eng_gam
        return np.interp(eng_gam, __e_gams, __spec_mu)
    else:
        return 0.0

@cython.cdivision(True)
cdef double __dnde_fsr_cp_vrf(double egam, double mv):
    cdef double mupi = mpi / mv
    cdef double x = 2. * egam / mv
    cdef double xmin = 0.0
    cdef double xmax = 1 - 4. * mupi**2
    cdef double dynamic
    cdef double coeff
    cdef double result

    if x < xmin or x > xmax:
        return 0.0

    coeff = qe**2 / (4. * (1 - 4 * mupi**2)**1.5 * M_PI**2)

    dynamic = ((2 * sqrt(1 - 4 * mupi**2 - x) *
                (-1 - 4 * mupi**2 * (-1 + x) + x + x**2)) / sqrt(1 - x) +
               (-1 + 4 * mupi**2) * (-1 + 2 * mupi**2 + x) *
               log((1 + sqrt(1 - x) * sqrt(1 - 4 * mupi**2 - x) - x)**2 /
                   (-1 + sqrt(1 - x) * sqrt(1 - 4 * mupi**2 - x) + x)**2)) / x

    result = dynamic * coeff

    return 2 * result / mv

@cython.cdivision(True)
cdef double __dnde_fsr_l_vrf(double egam, double ml, double mv):
    cdef double mul = ml / mv
    cdef double x = 2. * egam / mv

    cdef double xmin = 0.0
    cdef double xmax = 1 - 4. * mul**2

    cdef double dynamic
    cdef double coeff
    cdef double result

    if x < xmin or x > xmax:
        return 0.0

    coeff = -qe**2 / (8. * sqrt(1 - 4 * mul**2) * (1 + 2 * mul**2) * M_PI**2)

    dynamic = ((2 * sqrt(1 - 4 * mul**2 - x) *
                (2 - 4 * mul**2 * (-1 + x) - 2 * x + x**2)) / sqrt(1 - x) +
               (2 - 8 * mul**4 - 4 * mul**2 * x + (-2 + x) * x) * log(
                (-1 + sqrt(1 - x) * sqrt(1 - 4 * mul**2 - x) + x)**2 / (
                        1 + sqrt(1 - x) * sqrt(1 - 4 * mul**2 - x) - x)**2)) / x

    result = dynamic * coeff

    return 2 * result / mv

@cython.cdivision(True)
@cython.boundscheck(True)
@cython.wraparound(False)
cdef double __integrand(double cl, double eng_gam, double eng_v,
                        double mv, np.ndarray[double] pws, str mode):
    """
    Integrand of the boost integralself.

    Parameters
    ----------
    cl : float
        Angle the final state particle make with respect to the z-axis.
    eng_gams : float
        Gamma-ray energy to evaluate spectrum at.
    eng_s : float
        Energy of the scalar mediator.
    params :
        Scalar mediator model parameters.

    Returns
    -------
    integrand : float
        The value of the boost integral.
    """
    cdef double pwee = pws[0]
    cdef double pwmumu = pws[1]
    cdef double pwpi0g = pws[2]
    cdef double pwpipi = pws[3]

    cdef double beta = sqrt(1. - (mv / eng_v)**2)
    cdef double gamma = eng_v / mv

    cdef double jac = 1. / (2. * gamma * abs(1. - beta * cl))

    cdef double eng_gam_vrf = eng_gam * gamma * (1. - beta * cl)

    cdef double dnde = 0.0

    cdef double dnde_ee_f = pwee * __dnde_fsr_l_vrf(eng_gam_vrf, me, mv)
    cdef double dnde_mu_f = pwmumu * __dnde_fsr_l_vrf(eng_gam_vrf, mmu, mv)

    cdef double dnde_cp_f = pwpipi * __dnde_fsr_cp_vrf(eng_gam_vrf, mv)

    cdef double dnde_cp_d = 2. * pwpipi * __interp_spec(eng_gam_vrf, "cp")

    # Neutral pion energy is:
    cdef double e_pi0 = 0.5 * (mpi0**2 + mv**2) / mv
    cdef double dnde_np_d = pwpi0g * CSpectrumPoint(eng_gam_vrf, e_pi0)

    cdef double dnde_mu_d = 2. * pwmumu * __interp_spec(eng_gam_vrf, "mu")

    dnde = dnde_ee_f + dnde_mu_f + dnde_cp_f + \
           dnde_cp_d + dnde_np_d + dnde_mu_d

    if mode == "total":
        return jac * dnde
    if mode == "e e g":
        return jac * dnde_ee_f
    if mode == "pi pi g":
        return jac * dnde_cp_f
    if mode == "pi pi":
        return jac * dnde_cp_d
    if mode == "pi0 g":
        return jac * dnde_np_d
    if mode == "mu mu g":
        return jac * dnde_mu_f
    if mode == "mu mu":
        return jac * dnde_mu_d

@cython.boundscheck(True)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double __dnde_decay_v(double eng_gam, double eng_v, double mv,
                           np.ndarray[double] pws, str mode):
    """
    Unvectorized dnde_decay_s

    Compute the gamma ray spectrum from the decay of the scalar mediator.

    Parameters
    ----------
    eng_gams : float or array-like
        Float or array of gamma-ray energies to evaluate spectrum at.
    eng_s : float
        Energy of the scalar mediator.
    params :
        Scalar mediator model parameters.

    Returns
    -------
    dnde : float or array-like
        Values of dnde at gamma-ray energies `eng_gams`.
    """
    cdef double lines_contrib = 0.0
    cdef double beta = sqrt(1. - pow(mv / eng_v, 2.))
    cdef double gamma = eng_v / mv

    if eng_v < mv:
        return 0.

    cdef double eplus = eng_v * (1. + beta) / 2.0
    cdef double eminus = eng_v * (1. - beta) / 2.0
    cdef double result = 0.0

    if eminus <= eng_gam <= eplus:
        lines_contrib = pws[2] / (eng_v * beta)

    result = quad(__integrand, -1.0, 1.0, points=[-1.0, 1.0],
                  args=(eng_gam, eng_v, mv, pws, mode), epsabs=10**-10.,
                  epsrel=10**-5.)[0]

    if mode == "pi0 g" or mode == "total":
        return result + lines_contrib

    else:
        return result

@cython.boundscheck(True)
@cython.wraparound(False)
def dnde_decay_v_pt(double eng_gam, double eng_v, double mv,
                    np.ndarray[double] pws, str mode):
    """
    Compute the gamma ray spectrum from the decay of the scalar mediator.

    Parameters
    ----------
    eng_gams : float
        Gamma-ray energy to evaluate spectrum at.
    eng_s : float
        Energy of the scalar mediator.
    params :
        Scalar mediator model parameters.

    Returns
    -------
    dnde : float or array-like
        Value of dnde at gamma-ray energy `eng_gam`.
    """
    __set_spectra(mv)
    return __dnde_decay_v(eng_gam, eng_v, mv, pws, mode)

@cython.boundscheck(True)
@cython.wraparound(False)
def dnde_decay_v(np.ndarray[double] eng_gam, double eng_v, double mv,
                 np.ndarray[double] pws, str mode):
    """
    Compute the gamma ray spectrum from the decay of the scalar mediator.

    Parameters
    ----------
    eng_gams : float
        Gamma-ray energy to evaluate spectrum at.
    eng_s : float
        Energy of the scalar mediator.
    params :
        Scalar mediator model parameters.

    Returns
    -------
    dnde : float or array-like
        Value of dnde at gamma-ray energy `eng_gam`.
    """
    __set_spectra(mv)
    cdef int num_pts = len(eng_gam)
    cdef int i

    spec = np.zeros(num_pts, dtype=np.float64)

    for i in range(num_pts):
        spec[i] = __dnde_decay_v(eng_gam[i], eng_v, mv, pws, mode)

    return spec



from scipy.interpolate import interp1d
import numpy as np
from pkg_resources import resource_filename
from hazma.parameters import temp_cmb_formation

"""
Functions required for computing CMB limits and related quantities.
"""

# Get paths to files inside the module
f_eff_ep_rf = resource_filename(__name__, "cmb_data/f_eff_ep.dat")
f_eff_g_rf = resource_filename(__name__, "cmb_data/f_eff_g.dat")

# Load f_eff^{e+ e-}
f_eff_ep_data = np.loadtxt(f_eff_ep_rf, delimiter=",").T
f_eff_ep = interp1d(f_eff_ep_data[0] / 1.0e6, f_eff_ep_data[1])  # eV -> MeV

# Load f_eff^{e+ e-}
f_eff_g_data = np.loadtxt(f_eff_g_rf, delimiter=",").T
f_eff_g = interp1d(f_eff_g_data[0] / 1.0e6, f_eff_g_data[1])  # eV -> MeV

#: Planck 2018 95% upper limit on p_ann from temperature + polarization
#: measurements, in cm^3 s^-1 MeV^-1
p_ann_planck_temp_pol = 3.5e-31  # temperature + polarization
#: Planck 2018 95% upper limit on p_ann from temperature + polarization +
#: lensing measurements, in cm^3 s^-1 MeV^-1
p_ann_planck_temp_pol_lensing = 3.3e-31  # temp + pol + lensing
#: Planck 2018 95% upper limit on p_ann from temperature + polarization +
#: lensing + BAO measurements, in cm^3 s^-1 MeV^-1
p_ann_planck_temp_pol_lensing_bao = 3.2e-31  # temp + pol + lensing + BAO


def vx_cmb(mx, x_kd):
    """Computes the DM relative velocity at CMB using eq. 28 from `this
    reference <https://arxiv.org/abs/1309.4091>`_.

    Parameters
    ----------
    mx : float
        Dark matter mass in MeV.
    x_kd: float
        T_kd / m_x, where T_kd is the dark matter's kinetic decoupling
        temperature.

    Returns
    -------
    v_x : float
        The DM relative velocity at the time of CMB formation.
    """
    return 2.0e-4 * 10e6 * temp_cmb_formation / mx * np.sqrt(1.0e-4 / x_kd)



from scipy.interpolate import interp1d
from scipy.integrate import quad
from hazma.cmb import vx_cmb, f_eff_g, f_eff_ep, p_ann_planck_temp_pol

import numpy as np


class TheoryCMB:
    def cmb_limit(self, x_kd=1.0e-4, p_ann=p_ann_planck_temp_pol):
        r"""
        Computes the CMB limit on <sigma v>.

        This is derived by requiring that

        .. math::
            f_{\mathrm{eff}} \langle \sigma v \rangle / m_{\chi} < p_{\mathrm{ann}},

        where :math:`f_{\mathrm{eff}}` is the efficiency with which dark matter
        annihilations around recombination inject energy into the plasma and
        :math:`p_{\mathrm{ann}}` is derived from CMB observations.

        Parameters
        ----------
        x_kd: float
            T_kd / m_x, where T_kd is the dark matter's kinetic decoupling
            temperature. This will be computed self-consistently in future
            versions of ``hazma``.
        p_ann : float
            Constraint on energy release per DM annihilation in cm^3 s^-1
            MeV^-1.

        Returns
        -------
        <sigma v> : float
            Upper bound on <sigma v>, in cm^3 s^-1.
        """
        # TODO: account for non-self-conjugate DM. See discussion in Gondolo
        # and Gelmini.
        # if self.self_conjugate:
        #     factor = 1.0
        # else:
        #     factor = 0.5

        return p_ann * self.mx / self.f_eff(x_kd)

    def _f_eff_helper(self, fs, x_kd=1e-4, mode="quad"):
        """Computes f_eff^gg or f_eff^ep for DM annihilation.

        Parameters
        ----------
        fs : string
            "g g" or "e e", depending on which f_eff the user wants to compute.
        x_kd: float
            T_kd / m_x, where T_kd is the dark matter's kinetic decoupling
            temperature.
        mode : string
            "quad" or "interp". The first mode should be used if fs is "g g"
            ("e e") and none of the gamma ray spectrum functions (positron
            spectrum functions) use RAMBO as it is more accurate.

        Returns
        -------
        f_eff_dm : float
            f_eff for photons or electrons and positrons.
        """
        # Center of mass energy
        e_cm = 2.0 * self.mx * (1.0 + 0.5 * vx_cmb(self.mx, x_kd) ** 2)

        if fs == "g g":
            f_eff_base = f_eff_g
            lines = self.gamma_ray_lines(e_cm)
            spec_fn = self.total_spectrum
        elif fs == "e e":
            f_eff_base = f_eff_ep
            lines = self.positron_lines(e_cm)

            def spec_fn(es, e_cm):
                return 2.0 * self.total_positron_spectrum(es, e_cm)

        # Lower bound on integrals. Upper bound is many GeV, so we don't need
        # to do error checking.
        e_min = f_eff_base.x[0]

        # Continuum contributions from photons. Create an interpolator to avoid
        # recomputing spectrum.
        if mode == "interp":
            # If RAMBO is needed to compute the spectrum, it is prohibitively
            # time-consuming to try integrating the spectrum function. Instead,
            # simultaneously compute the spectrum over a grid of points.
            es = np.geomspace(e_min, e_cm / 2, 1000)
            dnde_tot = spec_fn(es, e_cm)
            spec_interp = interp1d(es, dnde_tot, bounds_error=False, fill_value=0.0)

            def integrand(e):
                return e * spec_interp(e) * f_eff_base(e)

            f_eff_dm = quad(integrand, e_min, e_cm / 2, epsabs=0, epsrel=1e-3)[0] / e_cm
        elif mode == "quad":
            # If RAMBO is not needed to compute the spectrum, this will give
            # much cleaner results.
            def integrand(e):
                return e * spec_fn(e, e_cm) * f_eff_base(e)

            f_eff_dm = quad(integrand, e_min, e_cm / 2, epsabs=0, epsrel=1e-3)[0] / e_cm

        # Sum up line contributions
        f_eff_line_dm = 0.0
        for ch, line in lines.items():
            energy = line["energy"]

            # Make sure the base f_eff is defined at this energy
            if energy > e_min:
                bf = line["bf"]
                multiplicity = 2.0 if ch == fs else 1.0
                f_eff_line_dm += energy * bf * f_eff_base(energy) * multiplicity / e_cm

        return f_eff_dm + f_eff_line_dm

    def f_eff_g(self, x_kd=1e-4):
        return self._f_eff_helper("g g", x_kd, "quad")

    def f_eff_ep(self, x_kd=1e-4):
        return self._f_eff_helper("e e", x_kd, "quad")

    def f_eff(self, x_kd=1.0e-4):
        """
        Computes :math:`f_{\mathrm{eff}}` the efficiency with which dark matter
        annihilations around recombination inject energy into the thermal
        plasma.
        """
        return self.f_eff_ep(x_kd) + self.f_eff_g(x_kd)



__all__ = [
    "background_model",
    "cmb",
    "constraint_parameters",
    "decay",
    "flux_measurement",
    "gamma_ray_parameters",
    "gamma_ray",
    "hazma_errors",
    "parameters",
    "positron_spectra",
    "rambo",
]



import numpy as np
cimport numpy as np
import cython
import warnings

ctypedef np.float64_t DBL_T


@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.ndarray[DBL_T, ndim=2] split_point(np.ndarray[DBL_T, ndim=1] l,
                                           int num_fsp):
    """
    Returns a list of four momentum from a flattened list.
    """
    cdef int i, j
    cdef np.ndarray[DBL_T, ndim=2] kList

    kList = np.empty((num_fsp, 4), dtype=np.float64)

    for i in range(num_fsp):
        for j in range(4):
            kList[i, j] = l[4 * i + j]
    return kList



@cython.boundscheck(False)
@cython.wraparound(False)
def apply_matrix_elem(np.ndarray[DBL_T, ndim=2] pts, int num_ps_pts,
                      int num_fsp, mat_elem_sqrd):
    """
    Applies the matrix element squared to the weights.
    """
    cdef int i
    cdef double mat_elem2

    for i in range(num_ps_pts):
        mat_elem2 = mat_elem_sqrd(split_point(pts[i], num_fsp))
        if mat_elem2 < 0:
            warnings.warn('Negative matrix element squared encountered...')
        pts[i, 4 * num_fsp] = pts[i, 4 * num_fsp] * mat_elem2
    return pts



class BackgroundModel(object):
    """
    Represents a gamma ray background model, which is required for computing
    projected limits for planned gamma-ray detectors.

    Parameters
    ----------
    e_range : [float, float]
        Minimum and maximum photon energies for which this model is valid, in
        MeV.
    dPhi_dEdOmega : np.array
        Background gamma ray flux (MeV^-1 sr^-1 m^-2 s^-1) as a function of
        photon energy (MeV). This function must be vectorized.
    """

    def __init__(self, e_range, dPhi_dEdOmega):
        self.e_range = e_range
        self.__dPhi_dEdOmega = dPhi_dEdOmega

    def dPhi_dEdOmega(self, es):
        """Computes this background model's gamma ray flux.

        Parameters
        ----------
        es : float or np.array
            Photon energy/energies at which to compute

        Returns
        -------
        dPhi_dEdOmega : np.array
            Background gamma ray flux, in MeV^-1 sr^-1 m^-2 s^-1. For any
            energies outside of ``self.e_range``, ``np.nan`` is returned.
        """
        if hasattr(es, "__len__"):
            # Check if any energies are out of bounds
            es_out_of_bounds = es[(es < self.e_range[0]) | (es > self.e_range[1])]

            if len(es_out_of_bounds) == 0:
                return self.__dPhi_dEdOmega(es)
            else:
                raise ValueError(
                    "The gamma ray background model is not "
                    "applicable for energy %f MeV." % es_out_of_bounds[0]
                )
        else:
            if es < self.e_range[0] or es > self.e_range[1]:
                raise ValueError(
                    "The gamma ray background model is not "
                    "applicable for energy %f MeV." % es
                )
            else:
                return self.__dPhi_dEdOmega(es)



from hazma.positron_helper_functions.positron_muon cimport CSpectrum as muspec
from scipy.integrate import quad
from libc.math cimport sqrt, pow, log10
import numpy as np
cimport numpy as np
import cython

include "parameters.pxd"

cdef double mmu = MASS_MU
cdef double mpi = MASS_PI
cdef double me = MASS_E

cdef double eng_mu_pi_rf = (mpi * mpi + mmu * mmu) / (2.0 * mpi)

cdef double beta_mu = sqrt(1.0 - pow(mmu / eng_mu_pi_rf, 2))
cdef double gamma_mu = eng_mu_pi_rf / mmu

cdef double eng_p_max_mu_rf = (me * me + mmu * mmu) / (2.0 * mmu)
cdef double eng_p_max_pi_rf = eng_p_max_mu_rf * gamma_mu * (1.0 + beta_mu)

# TODO: Really?? 10000 pts?
eng_ps_mu = np.logspace(log10(me), log10(eng_p_max_pi_rf), num=500, dtype=np.float64)

cdef np.ndarray __muspec = muspec(eng_ps_mu, eng_mu_pi_rf)

cdef double __muon_spectrum(double eng_p):
    """
    Returns the muon spectrum in the pion rest frame.

    Parameters
    ----------
    eng_p : double
        Energy of the positron.

    Returns
    -------
    dnde_mu : double
        Value of muon positron spectrum at an positron energy of `eng_p`
    """
    return np.interp(eng_p, eng_ps_mu, __muspec)

@cython.cdivision(True)
cdef double __integrand(double cl, double eng_p, double eng_pi):
    """
    Returns the integrand of the boost integral at a given angle.

    Parameters
    ----------
    eng_p : double
        Energy of the positron.
    eng_pi : double
        Energy of the pion.

    Returns
    -------
    integrand : double
        Integrand of the boost integral at angle `cl`, positron energy `eng_p`
        and pion energy `eng_pi`.
    """
    if eng_p < me:
        return 0.0
    cdef double p = sqrt(eng_p * eng_p - me * me)
    cdef double gamma = eng_pi / mpi
    cdef double beta = sqrt(1.0 - pow(mpi / eng_pi, 2))
    cdef double jac = p / (2. * sqrt((1 + pow(beta * cl, 2)) * eng_p * eng_p -
                                     (1 + beta * beta * (-1 + cl * cl)) *
                                     me * me -
                                     2 * beta * cl * eng_p * p) * gamma)

    cdef double eng_p_pi_rf = gamma * (eng_p - p * beta * cl)

    return BR_PI_TO_MUNU * jac * __muon_spectrum(eng_p_pi_rf)

cdef double CSpectrumPoint(double eng_p, double eng_pi):
    """
    Cythonized version of SpectrumPoint.

    Returns the positron spectrum from a charged pion for a single positron
    energy.

    Parameters
    ----------
    eng_p : double
        Energy of the positron.
    eng_pi : double
        Energy of the pion.

    Returns
    -------
    dnde_cpi : double
        Positron spectrum from a charged pion given positron energy `eng_p`
        and pion energy `eng_pi`.
    """
    if eng_pi < mpi:
        return 0.0

    return quad(__integrand, -1.0, 1.0, points=[-1.0, 1.0],
                args=(eng_p, eng_pi), epsabs=1e-10, epsrel=1e-4)[0]

@cython.boundscheck(True)
@cython.wraparound(False)
cdef np.ndarray CSpectrum(np.ndarray eng_ps, double eng_pi):
    """
    Cythonized version of Spectrum.

    Returns the positron spectrum from a charged pion for many positron
    energies.

    Parameters
    ----------
    eng_ps : numpy.array
        Energies of the positron.
    eng_pi : double
        Energy of the pion.

    Returns
    -------
    dnde_cpi : double
        Positron spectrum from a charged pion given positron energies `engs_p`
        and pion energy `eng_pi`.
    """
    cdef int num_pts = len(eng_ps)
    cdef np.ndarray spec = np.zeros(num_pts, dtype=np.float64)
    cdef int i = 0

    for i in range(num_pts):
        spec[i] = CSpectrumPoint(eng_ps[i], eng_pi)

    return spec

def SpectrumPoint(double eng_p, double eng_pi):
    """
    Returns the positron spectrum from a charged pion for a single positron
    energy.

    Parameters
    ----------
    eng_p : double
        Energy of the positron.
    eng_pi : double
        Energy of the pion.

    Returns
    -------
    dnde_cpi : double
        Positron spectrum from a charged pion given positron energy `eng_p`
        and pion energy `eng_pi`.
    """
    return CSpectrum(eng_p, eng_pi)

@cython.boundscheck(True)
@cython.wraparound(False)
def Spectrum(np.ndarray eng_ps, double eng_pi):
    """
    Returns the positron spectrum from a charged pion for many positron
    energies.

    Parameters
    ----------
    eng_ps : numpy.array
        Energies of the positron.
    eng_pi : double
        Energy of the pion.

    Returns
    -------
    dnde_cpi : double
        Positron spectrum from a charged pion given positron energies `engs_p`
        and pion energy `eng_pi`.
    """
    return CSpectrum(eng_ps, eng_pi)



import numpy as np


class FluxMeasurement:
    """
    Container for all information about a completed gamma ray analysis.

    Attributes
    ----------
    e_lows : np.array
        Lower edges of energy bins.
    e_highs : np.array
        Upper edges of energy bins.
    fluxes : np.array
        Flux measurements for each bin (MeV^-1 cm^-2 s^-1 sr^-1).
    upper_errors : np.array
        Size of upper error bars on flux measurements (MeV^-1 cm^-2 s^-1
        sr^-1).
    lower_errors : np.array
        Size of lower error bars on flux measurements (MeV^-1 cm^-2 s^-1
        sr^-1).
    energy_res : callable
        Function returning energy resolution (Delta E / E) as a function of
        photon energy.
    target : TargetParams
        Information about the target observed for this measurement.
    """

    def __init__(self, obs_rf, energy_res, target, power=2):
        """Constructor.

        Parameters
        ----------
        obs_rf : str
            Name of file containing observation information. The columns of
            this file must be:

                1. Lower bin edge (MeV)
                2. Upper bin edge (MeV)
                3. :math:`E^2 d^2 \Phi/dE d\Omega` (MeV cm^-2 s^-1 sr^-1)
                4. Upper error bar (MeV cm^-2 s^-1 sr^-1)
                5. Lower error bar (MeV cm^-2 s^-1 sr^-1)

            Note that the error bar values are their y-coordinates, not their
            relative distances from the central flux.
        energy_res : callable
            Energy resolution function.
        target : TargetParams
            The target of the analysis
        """
        self.e_lows, self.e_highs, self.fluxes, self.upper_errors, self.lower_errors = np.loadtxt(
            obs_rf, delimiter=","
        ).T

        # Get bin central values
        self._e_bins = 0.5 * (self.e_lows + self.e_highs)

        # E^2 dN/dE -> dN/dE
        self.fluxes /= self._e_bins ** power

        # Compute upper and lower error bars
        self.upper_errors = self.upper_errors / self._e_bins ** power - self.fluxes
        self.lower_errors = self.fluxes - self.lower_errors / self._e_bins ** power

        # Load energy resolution
        self.energy_res = energy_res

        # Store analysis region
        self.target = target



import numpy as np
from hazma.parameters import (
    charged_B_mass as mB,
    charged_kaon_mass as mk,
    charged_pion_mass as mpi,
    down_quark_mass as mdq,
    bottom_quark_mass as mbq,
    strange_quark_mass as msq,
    long_kaon_mass as mkl,
    top_quark_mass as mtq,
    neutral_pion_mass as mpi0,
    GF,
    vh,
    Vts,
    Vtb,
    Vtd,
)
from hazma.constraint_parameters import (
    B_k_invis_obs,
    k_pi_invis_obs,
    kl_pi0_mu_mu_obs,
    kl_pi0_e_e_obs,
    B_k_mu_mu_obs,
    B_k_e_e_obs,
    k_width,
    kl_width,
    B_width,
    cm_to_inv_MeV,
    bd_charm,
)

# TODO
# * 3-body phase space integration for s -> x x contribution to invisible
#   decays


class ScalarMediatorConstraints:
    def _lambda_ps(self, a, b, c):
        """Phase space factor."""
        return (a - b - c) ** 2 - 4.0 * b * c

    def width_B_k_s(self):
        if self.ms <= mB - mk:
            f0B = 0.33 / (1.0 - self.ms ** 2 / (38.0e3) ** 2)
            # FCNC couplings
            hSsb = (
                3.0
                * np.sqrt(2)
                * (msq + mbq)
                * GF
                * mtq ** 2
                * Vts.conjugate()
                * Vtb
                / (32.0 * np.pi ** 2 * vh)
            )

            return (
                1.0
                / (16.0 * np.pi * mB ** 3)
                * ((mB ** 2 - mk ** 2) / (mbq - msq)) ** 2
                * f0B ** 2
                * abs(hSsb) ** 2
                * self.gsff ** 2
                * np.sqrt(self._lambda_ps(mB ** 2, mk ** 2, self.ms ** 2))
            )
        else:
            return 0.0

    def constraint_B_k_invis(self):
        """Constraint function for B+ -> K+ invis."""
        ms = self.ms
        width_contr = 0.0

        # Make sure scalar mass doesn't fall outside of kinematic bounds
        if np.any([s[0] <= ms ** 2 <= s[1] for s in B_k_invis_obs.s_bounds]):
            widths_s = self.partial_widths()
            width_s = widths_s["total"]
            width_s_sm = width_s - widths_s["x x"]  # Gamma_{S->SM}

            # Magnitude of S' 3-momentum
            ps = np.sqrt(
                (mB - mk - ms) * (mB + mk - ms) * (mB - mk + ms) * (mB + mk + ms)
            ) / (2.0 * mB)
            # Probability that S decays outside the detector
            pr_invis = np.exp(-B_k_invis_obs.r_max * cm_to_inv_MeV * width_s * ms / ps)

            # Compute the total contribution to the invisible decay width
            width_contr = (
                self.width_B_k_s() * (widths_s["x x"] + pr_invis * width_s_sm) / width_s
            )

        return B_k_invis_obs.width_bound - width_contr

    def constraint_B_k_mu_mu(self):
        """Constraint function for B+ -> K+ mu+ mu-."""
        ms = self.ms
        width_contr = 0.0

        # Make sure scalar mass doesn't fall outside of kinematic bounds
        if np.any([s[0] <= ms ** 2 <= s[1] for s in B_k_mu_mu_obs.s_bounds]):
            widths_s = self.partial_widths()
            width_s = widths_s["total"]

            # Magnitude of S' 3-momentum
            ps = np.sqrt(
                (mB - mk - ms) * (mB + mk - ms) * (mB - mk + ms) * (mB + mk + ms)
            ) / (2.0 * mB)
            # Probability that S decays close to the primary vertex
            pr_vis = 1.0 - np.exp(
                -B_k_mu_mu_obs.r_max * cm_to_inv_MeV * width_s * ms / ps
            )

            # print(pr_vis)
            # print(widths_s["mu mu"] / width_s)

            # Compute the contribution to the mu mu decay width
            width_contr = self.width_B_k_s() * widths_s["mu mu"] / width_s * pr_vis

        return B_k_mu_mu_obs.width_bound - width_contr

    def constraint_B_k_e_e(self):
        """Constraint function for B+ -> K+ e+ e-."""
        ms = self.ms
        width_contr = 0.0

        # Make sure scalar mass doesn't fall outside of kinematic bounds
        if np.any([s[0] <= ms ** 2 <= s[1] for s in B_k_e_e_obs.s_bounds]):
            widths_s = self.partial_widths()
            width_s = widths_s["total"]

            # Magnitude of S' 3-momentum
            ps = np.sqrt(
                (mB - mk - ms) * (mB + mk - ms) * (mB - mk + ms) * (mB + mk + ms)
            ) / (2.0 * mB)
            # Probability that S decays close to the primary vertex
            pr_vis = 1.0 - np.exp(
                -B_k_e_e_obs.r_max * cm_to_inv_MeV * width_s * ms / ps
            )

            # Compute the contribution to the e e decay width
            width_contr = self.width_B_k_s() * widths_s["e e"] / width_s * pr_vis

        return B_k_e_e_obs.width_bound - width_contr

    def width_k_pi_s(self):
        if self.ms < mk - mpi:
            # FCNC couplings
            hSsd = (
                3.0
                * np.sqrt(2)
                * (msq + mdq)
                * GF
                * mtq ** 2
                * Vts.conjugate()
                * Vtd
                / (32.0 * np.pi ** 2 * vh)
            )

            return (
                1.0
                / (16.0 * np.pi * mk ** 3)
                * ((mk ** 2 - mpi ** 2) / (msq - mdq)) ** 2
                * abs(hSsd) ** 2
                * self.gsff ** 2
                * np.sqrt(self._lambda_ps(mk ** 2, mpi ** 2, self.ms ** 2))
            )
        else:
            return 0.0

    def constraint_k_pi_invis(self):
        """Constraint function for K+ -> pi+ invis."""
        width_contr = 0.0
        ms = self.ms

        # Make sure scalar mass doesn't fall outside of kinematic bounds
        if np.any([s[0] <= ms ** 2 <= s[1] for s in k_pi_invis_obs.s_bounds]):
            widths_s = self.partial_widths()
            width_s = widths_s["total"]
            width_s_sm = width_s - widths_s["x x"]  # Gamma_{S->SM}

            # Magnitude of S' 3-momentum
            ps = np.sqrt(
                (mk - mpi - ms) * (mk + mpi - ms) * (mk - mpi + ms) * (mk + mpi + ms)
            ) / (2.0 * mk)
            # Probability that S decays outside the detector
            pr_invis = np.exp(-k_pi_invis_obs.r_max * cm_to_inv_MeV * width_s * ms / ps)

            # Compute the total contribution to the invisible decay width
            width_contr = (
                self.width_k_pi_s()
                * (widths_s["x x"] + pr_invis * width_s_sm)
                / width_s
            )

        return k_pi_invis_obs.width_bound - width_contr

    def width_kl_pi0_s(self):
        if self.ms < mkl - mpi0:
            # FCNC couplings
            hSsd = (
                3.0
                * np.sqrt(2)
                * (msq + mdq)
                * GF
                * mtq ** 2
                * Vts.conjugate()
                * Vtd
                / (32.0 * np.pi ** 2 * vh)
            )

            return (
                1.0
                / (16.0 * np.pi * mkl ** 3)
                * ((mkl ** 2 - mpi0 ** 2) / (msq - mdq)) ** 2
                * hSsd.imag ** 2
                * self.gsff ** 2
                * np.sqrt(self._lambda_ps(mkl ** 2, mpi0 ** 2, self.ms ** 2))
            )
        else:
            return 0.0

    def constraint_kl_pi0_mu_mu(self):
        """Constraint function for K_L -> pi0 mu+ mu-."""
        width_contr = 0.0
        ms = self.ms

        # Make sure scalar mass doesn't fall outside of kinematic bounds
        if np.any([s[0] <= ms ** 2 <= s[1] for s in kl_pi0_mu_mu_obs.s_bounds]):
            widths_s = self.partial_widths()
            width_s = widths_s["total"]

            # Magnitude of S' 3-momentum
            ps = np.sqrt(
                (mkl - mpi0 - ms)
                * (mkl + mpi0 - ms)
                * (mkl - mpi0 + ms)
                * (mkl + mpi0 + ms)
            ) / (2.0 * mkl)
            # Probability that S decays close to the primary vertex
            pr_vis = 1.0 - np.exp(
                -kl_pi0_mu_mu_obs.r_max * cm_to_inv_MeV * width_s * ms / ps
            )

            # Compute the contribution to the mu mu decay width
            width_contr = self.width_kl_pi0_s() * widths_s["mu mu"] / width_s * pr_vis

        return kl_pi0_mu_mu_obs.width_bound - width_contr

    def constraint_kl_pi0_e_e(self):
        """Constraint function for K_L -> pi0 e+ e-."""
        width_contr = 0.0
        ms = self.ms

        # Make sure scalar mass doesn't fall outside of kinematic bounds
        if np.any([s[0] <= ms ** 2 <= s[1] for s in kl_pi0_e_e_obs.s_bounds]):
            widths_s = self.partial_widths()
            width_s = widths_s["total"]

            # Magnitude of S' 3-momentum
            ps = np.sqrt(
                (mkl - mpi0 - ms)
                * (mkl + mpi0 - ms)
                * (mkl - mpi0 + ms)
                * (mkl + mpi0 + ms)
            ) / (2.0 * mkl)
            # Probability that S decays close to the primary vertex
            pr_vis = 1.0 - np.exp(
                -kl_pi0_e_e_obs.r_max * cm_to_inv_MeV * width_s * ms / ps
            )
            # Compute the contribution to the e e decay width
            width_contr = self.width_kl_pi0_s() * widths_s["e e"] / width_s * pr_vis

        return kl_pi0_e_e_obs.width_bound - width_contr

    def width_B_xs_s(self):
        # B -> X_s S, where X_s is any strange meson
        if self.ms <= mB - mk:  # since K^+ is the lightest strange meson
            hSsb = (
                3.0
                * np.sqrt(2)
                * (msq + mbq)
                * GF
                * mtq ** 2
                * Vts.conjugate()
                * Vtb
                / (32.0 * np.pi ** 2 * vh)
            )

            val = (
                1.0
                / (8.0 * np.pi)
                * (mbq ** 2 - self.ms ** 2) ** 2
                / mbq ** 3
                * abs(hSsb) ** 2
                * self.gsff ** 2
            )

            return val
        else:
            return 0.0

    def constrain_beam_dump(self, bd_params=bd_charm):
        # Branching fractions for production in meson decays
        br_k_pi_s = self.width_k_pi_s() / k_width
        br_kl_pi0_s = self.width_kl_pi0_s() / kl_width
        br_B_xs_s = self.width_B_xs_s() / B_width

        assert br_k_pi_s >= 0
        assert br_kl_pi0_s >= 0
        assert br_B_xs_s >= 0

        # Probability that collision produces S via meson decays
        br_pN_s = (
            bd_params.br_pN_k * (0.5 * br_k_pi_s + 0.25 * br_kl_pi0_s)
            + bd_params.br_pN_B * br_B_xs_s
        )

        # Probability that S decays in the detector
        conv_fact = 1.0e2 * cm_to_inv_MeV  # converts m -> MeV^-1
        d = conv_fact * bd_params.dist
        L = conv_fact * bd_params.length

        gamma_beta = np.sqrt((bd_params.mediator_energy / self.ms) ** 2 - 1.0)
        width_s = self.width_s

        pr_decay_in_det = np.exp(-d * width_s / gamma_beta) * (
            1.0 - np.exp(-L * width_s / gamma_beta)
        )

        # Probability that S decays into a visible final state
        br_visible = (
            np.sum(
                [
                    bf
                    for fs, bf in self.partial_widths().items()
                    if fs in bd_params.visible_fss
                ]
            )
            / width_s
        )

        # Number of expected events
        n_dec_th = bd_params.n_pot * br_pN_s * pr_decay_in_det * br_visible

        assert n_dec_th >= 0

        return bd_params.n_dec - n_dec_th

    def constraints(self):
        return {
            "B -> k invis": self.constraint_B_k_invis,
            "B -> k mu mu": self.constraint_B_k_mu_mu,
            # "B -> k e e": self.constraint_B_k_e_e,  # worthlessly weak!
            "k -> pi invis": self.constraint_k_pi_invis,
            "kl -> pi0 mu mu": self.constraint_kl_pi0_mu_mu,
            "kl -> pi0 e e": self.constraint_kl_pi0_e_e,
            "CHARM": self.constrain_beam_dump,
        }



"""
Module for computing decay spectra from a muon and light mesons.

@author: Logan Morrison and Adam Coogan
@date: January 2018

"""
import numpy as np
from hazma.decay_helper_functions import decay_long_kaon
from hazma.decay_helper_functions import decay_charged_pion
from hazma.decay_helper_functions import decay_charged_kaon
from hazma.decay_helper_functions import decay_muon
from hazma.decay_helper_functions import decay_neutral_pion
from hazma.decay_helper_functions import decay_short_kaon


def muon(photon_energies, muon_energy):
    r"""Compute gamma-ray decay spectrum from the muon decay
    :math:`\mu^{\pm} \to e^{\pm} \nu_{e} \nu_{\mu}`.

    Parameters
    ----------
    photon_energies : float or numpy.ndarray
        Photon energy(ies) in laboratory frame.
    muon_energy : double
        Muon energy in laboratory frame.

    Returns
    -------
    spec : numpy.ndarray
        List of gamma ray spectrum values, dNdE, evaluated at
        ``photon_energies`` given muon energy ``eng_mu``.

    Examples
    --------

    Calculate spectrum for single gamma ray energy::

        from hazma import decay
        photon_energy, muon_energy = 200., 1000.
        decay.muon(photon_energy, muon_energy)

    Calculate spectrum for array of gamma ray energies::

        from hazma import decay
        import numpy as np
        photon_energies = np.logspace(0.0, 3.0, num=200, dtype=float)
        muon_energy = 1000.
        decay.muon(photon_energies, muon_energy)
    """
    # mu = decay_muon.Muon()
    if hasattr(photon_energies, "__len__"):
        return decay_muon.Spectrum(photon_energies, muon_energy)
    return decay_muon.SpectrumPoint(photon_energies, muon_energy)


def neutral_pion(photon_energies, pion_energy):
    r"""Compute gamma-ray spectrum from the neutral pion decay
    :math:`\pi^{0} \to \gamma \gamma`.

    Parameters
    ----------
    photon_energies : float or numpy.ndarray
        Photon energy(ies) in laboratory frame.
    pion_energy : float
        Neutral pion energy in laboratory frame.

    Returns
    -------
    spec : np.ndarray
        List of gamma ray spectrum values, dNdE, evaluated at
        `photon_energies` given neutral pion energy `pion_energy`.

    Examples
    --------

    Calculate spectrum for single gamma ray energy::

        from hazma import decay
        photon_energies, pion_energy = 200., 1000.
        decay.neutral_pion(photon_energies, pion_energy)

    Calculate spectrum for array of gamma ray energies::

        from hazma import decay
        import numpy as np
        photon_energies = np.logspace(0.0, 3.0, num=200, dtype=float)
        pion_energy = 1000.
        decay.neutral_pion(photon_energies, pion_energy)
    """
    if hasattr(photon_energies, "__len__"):
        return decay_neutral_pion.Spectrum(photon_energies, pion_energy)
    return decay_neutral_pion.SpectrumPoint(photon_energies, pion_energy)


def charged_pion(photon_energies, pion_energy, mode="total"):
    r"""Compute gamma-ray spectrum from the charged pion decay :math:`\pi^{\pm}
    \to \mu^{\pm} \nu_{\mu} \to e^{\pm} \nu_{e} \nu_{\mu} \gamma`.

    Parameters
    ----------
    photon_energies : float or numpy.ndarray
        Photon energy(ies) in laboratory frame.
    pion_energy : double
        Charged pion energy in laboratory frame.
    mode : str {"total"}
        The mode the user would like to have returned. The options are "total",
        "munu", "munug" and "enug".

    Returns
    -------
    spec : double np.ndarray
        List of gamma ray spectrum values, dNdE, evaluated at `photon_energies`
        given charged pion energy `eng_pi`.

    Examples
    --------
    Calculate spectrum for single gamma ray energy::

        from hazma import decay
        photon_energies, pion_energy = 200., 1000.
        decay.charged_pion(photon_energies, pion_energy)

    Calculate spectrum for array of gamma ray energies::

        from hazma import decay
        import numpy as np
        photon_energies = np.logspace(0.0, 3.0, num=200, dtype=float)
        pion_energy = 1000.
        decay.charged_pion(photon_energies, pion_energy)
    """

    if mode != "total" and mode != "munu" and mode != "munug" and mode != "enug":
        val_err_mess = (
            "mode '{}'' is not availible. Please use 'total'"
            + "'munu', 'munug' or 'enug'.".format(mode)
        )
        raise ValueError(val_err_mess)

    if hasattr(photon_energies, "__len__"):
        return decay_charged_pion.Spectrum(photon_energies, pion_energy, mode)
    return decay_charged_pion.SpectrumPoint(photon_energies, pion_energy, mode)


def charged_kaon(photon_energies, kaon_energy, mode="total"):
    r"""Compute gamma-ray spectrum from charged kaon decay into various final states.

    Parameters
    ----------
    photon_energies : float or numpy.ndarray
        Photon energy(ies) in laboratory frame.
    kaon_energy : float
        Charged kaon energy in laboratory frame.
    mode : str {"total"}
        The mode the user would like to have returned. The options are "total",
        "0enu", "0munu", "00p", "mmug", "munu", "p0", "p0g" and "ppm". Here
        "p" stands for pi plus, "m" stands for pi minus and "0" stands pi 0.

    Returns
    -------
    spec : np.ndarray
        List of gamma ray spectrum values, dNdE, evaluated at
        ``photon_energies`` given Kaon energy ``kaon_energy``.

    Notes
    -----
    The decay modes implemented are

    .. math:: K^{\pm} \to \mu^{\pm} \nu_{\mu}

    .. math:: K^{\pm} \to \pi^{\pm} \pi^{0}

    .. math:: K^{\pm} \to \pi^{\pm} \pi^{\mp} + \pi^{\pm}

    .. math:: K^{\pm} \to e^{\pm} \nu_{e}

    .. math:: K^{\pm} \to \mu^{\pm} \nu_{\mu} \pi^{0}

    .. math:: K^{\pm} \to \pi^{\pm} \pi^{0} \pi^{0}

    Examples
    --------

    Calculate spectrum for single gamma ray energy::

        from hazma import decay
        photon_energies, kaon_energy = 200., 1000.
        decay.charged_kaon(photon_energies, kaon_energy)

    Calculate spectrum for array of gamma ray energies::

        from hazma import decay
        import numpy as np
        photon_energies = np.logspace(0.0, 3.0, num=200, dtype=float)
        kaon_energy = 1000.
        decay.charged_kaon(photon_energies, kaon_energy)
    """

    if (
        mode != "total"
        and mode != "0enu"
        and mode != "0munu"
        and mode != "00p"
        and mode != "mmug"
        and mode != "munu"
        and mode != "p0"
        and mode != "p0g"
        and mode != "ppm"
    ):
        val_err_mess = (
            "mode '{}'' is not availible. Please use 'total'"
            + "'0enu', '0munu', '00p', 'mmug', 'munu', 'p0',"
            + " 'p0g' or 'ppm'.".format(mode)
        )
        raise ValueError(val_err_mess)

    if hasattr(photon_energies, "__len__"):
        return decay_charged_kaon.Spectrum(photon_energies, kaon_energy, mode)
    return decay_charged_kaon.SpectrumPoint(photon_energies, kaon_energy, mode)


def short_kaon(photon_energies, kaon_energy, mode="total"):
    r"""Compute gamma-ray spectrum from short kaon decay into various final states.

    Parameters
    ----------
    photon_energies : float or numpy.ndarray
        Photon energy(ies) in laboratory frame.
    kaon_energy : float
        Charged kaon energy in laboratory frame.
    mode : str
        The mode the user would like to have returned. The options are "total",
        "00", "pm" or "pmg". Here "p" stands for pi plus, "m" stands for pi
        minus and "0" stands pi 0.

    Returns
    -------
    spec : np.ndarray
        List of gamma ray spectrum values, dNdE, evaluated at
        ``photon_energies`` given muon energy ``eng_mu``.

    Notes
    -----
    The decay modes implemented are

    .. math:: K_{S} \to \pi^{+}  + \pi^{-}

    .. math:: K_{S} \to \pi^{0} + \pi^{0}

    Examples
    --------

    Calculate spectrum for single gamma ray energy::

        from hazma import decay
        photon_energies, kaon_energy = 200., 1000.
        decay.short_kaon(photon_energies, kaon_energy)

    Calculate spectrum for array of gamma ray energies::

        from hazma import decay
        import numpy as np
        photon_energies = np.logspace(0.0, 3.0, num=200, dtype=float)
        kaon_energy = 1000.
        decay.short_kaon(photon_energies, kaon_energy)
    """
    if mode != "total" and mode != "00" and mode != "pm" and mode != "pmg":
        val_err_mess = (
            "mode '{}'' is not available. Please use 'total'"
            + "'00', 'pm' or 'pmg'.".format(mode)
        )
        raise ValueError(val_err_mess)

    if hasattr(photon_energies, "__len__"):
        return decay_short_kaon.Spectrum(photon_energies, kaon_energy, mode)
    return decay_short_kaon.SpectrumPoint(photon_energies, kaon_energy, mode)


def long_kaon(photon_energies, kaon_energy, mode="total"):
    r"""Compute gamma-ray spectrum from long kaon decay into various final
    states.

    Parameters
    ----------
    photon_energies : float or numpy.ndarray
        Photon energy(ies) in laboratory frame.
    kaon_energy : float
        Charged kaon energy in laboratory frame.
    mode : str
        The mode the user would like to have returned. The options are "total",
        "000", "penu", "penug", "pm0", "pm0g", "pmunu" or "pmunug". Here "p"
        stands for pi plus, "m" stands for pi minus and "0" stands pi 0.

    Returns
    -------
    spec : np.ndarray
        List of gamma ray spectrum values, dNdE, evaluated at `photon_energies`
        given muon energy `kaon_energy`.

    Examples
    --------
    Calculate spectrum for single gamma ray energy::

        from hazma import decay
        photon_energies, kaon_energy = 200., 1000.
        decay.long_kaon(photon_energies, kaon_energy)

    Calculate spectrum for array of gamma ray energies::

        from hazma import decay
        import numpy as np
        photon_energies = np.logspace(0.0, 3.0, num=200, dtype=float)
        kaon_energy = 1000.
        decay.long_kaon(photon_energies, kaon_energy)

    Notes
    -----
    The decay modes implemented are

    .. math:: K_{L} \to \pi^{\pm} e^{\pm} \nu_{e}

    .. math:: K_{L} \to \pi^{\pm} \mu^{\mp} \nu_{\mu}

    .. math:: K_{L} \to \pi^{0} \pi^{0} \pi^{0}

    .. math:: K_{L} \to \pi^{+} \pi^{-} \pi^{0}

    """
    if (
        mode != "total"
        and mode != "000"
        and mode != "penu"
        and mode != "penug"
        and mode != "pm0"
        and mode != "pm0g"
        and mode != "pmunu"
        and mode != "pmunug"
    ):
        val_err_mess = (
            "mode '{}'' is not availible. Please use 'total'"
            + "'000', 'penu', 'penug', 'pm0', 'pm0g',"
            + "'pmunu' or 'pmunug'.".format(mode)
        )
        raise ValueError(val_err_mess)

    if hasattr(photon_energies, "__len__"):
        return decay_long_kaon.Spectrum(photon_energies, kaon_energy, mode)
    return decay_long_kaon.SpectrumPoint(photon_energies, kaon_energy, mode)


def electron(photon_energies, electron_energy):
    r"""Compute gamma-ray spectrum from electron decay (returns zero).

    The purpose of this function is so we can use the electron as a final
    state in `hazma.gamma_ray`.

    Parameters
    ----------
    photon_energies : float or numpy.ndarray
        Photon energy(ies) in laboratory frame.
    electron_energy : double
        Electron energy in laboratory frame.

    Returns
    -------
    spec : numpy.ndarray
        An array of zeros.
    """
    if hasattr(photon_energies, "__len__"):
        return np.array([0.0 for _ in photon_energies])
    return 0.0



import numpy as np
from scipy.integrate import trapz
from scipy.interpolate import InterpolatedUnivariateSpline, interp1d

"""
Physics constants and utility functions.
"""

# Masses (MeV)
electron_mass = 0.510998928  # electron
muon_mass = 105.6583715  # muon
neutral_pion_mass = 134.9766  # neutral pion
charged_pion_mass = 139.57018  # charged pion
neutral_kaon_mass = 497.61  # neutral kaon
long_kaon_mass = 497.614
charged_kaon_mass = 493.68  # charged Kaon
eta_mass = 547.86  # eta
eta_prime_mass = 957.8  # eta prime
rho_mass = 775.3  # rho
omega_mass = 782.7  # omega
charged_B_mass = 5279.29  # B^+ meson
pion_mass_chiral_limit = (neutral_pion_mass + charged_pion_mass) / 2.0
kaon_mass_chiral_limit = (neutral_kaon_mass + charged_kaon_mass) / 2.0

# Quark masses in MS-bar scheme
up_quark_mass = 2.3
down_quark_mass = 4.8
strange_quark_mass = 95.0
charm_quark_mass = 1.275e3
bottom_quark_mass = 4.18e3
top_quark_mass = 160.0e3

# Convert <sigma v> from MeV^2
cm_to_inv_MeV = 5.08e10  # MeV^-1 cm^-1
sv_inv_MeV_to_cm3_per_s = 1.0 / cm_to_inv_MeV ** 2 * 3e10  # cm^3/s * MeV^2

# Miscellaneous
alpha_em = 1.0 / 137.04  # fine structure constant.
GF = 1.1663787e-11  # Fermi constant in MeV**-2
vh = 246.22795e3  # Higgs VEV in MeV
qe = np.sqrt(4.0 * np.pi * alpha_em)
temp_cmb_formation = 0.235e-6  # CMB temperature at formation in MeV

# Charges
Qu = 2.0 / 3.0
Qd = -1.0 / 3.0
Qe = -1.0

# Low Energy constants
fpi0 = 91.924  # Neutral pion decay constant
fpi = 92.2138  # Charged pion decay constant
fk = 110.379  # Charged kaon decay constant
b0 = pion_mass_chiral_limit ** 2 / (up_quark_mass + down_quark_mass)
G8 = 5.47
G27 = 0.392
gv = 67.0
fv = 153.0

# The following low energy constants are for NLO ChiPT, evaluated at mu = mrho.
nlo_lec_mu = rho_mass
Lr1 = 0.56 * 1.0e-3
Lr2 = 1.21 * 1.0e-3
L3 = -2.79 * 1.0e-3
Lr4 = -0.36 * 1.0e-3
Lr5 = 1.4 * 1.0e-3
Lr6 = 0.07 * 1.0e-3
L7 = -0.44 * 1.0e-3
Lr8 = 0.78 * 1.0e-3

# SU(2) LECs
Er = 0.029
Gr = 0.0073

LECS = {
    "1": Lr1,
    "2": Lr2,
    "3": L3,
    "4": Lr4,
    "5": Lr5,
    "6": Lr6,
    "7": L7,
    "8": Lr8,
    "SU2_Er": Er,
    "SU2_Gr": Gr,
}

# SU(2) LECs
Er = 0.029
Gr = 0.0073

# CKM matrix elements
Vud = 0.974267
Vus = 0.2248
Vts = -0.0405 - 0.00075987j
Vtb = 0.999139
Vtd = 0.00823123 - 0.00328487j


# Widths (MeV)
rho_width = 146.2
B_width = 4.35e-10
k_width = 5.32e-14
kl_width = 1.29e-14


def convert_sigmav(sv, target):
    """Changes the units of <sigma v>.

    Parameters
    ----------
    sv : float
        Cross section in units of MeV^-2 or cm^3 / s.
    target : string
        Units to convert to. Must be "MeV^-2" or "cm^3/s" -- whichever units sv
        is NOT in.

    Returns
    -------
    sv : float
        sv converted to be in the target units.
    """
    # hbar^2 c^3 in units of MeV^2 cm^3 / s
    hbar2_c3 = (3.0e10) ** 3 * (6.58e-22) ** 2

    if target == "cm^3 / s":
        return sv * hbar2_c3
    elif target == "MeV^-2":
        return sv / hbar2_c3


def load_interp(rf_name, bounds_error=False, fill_value=0.0):
    """Creates an interpolator from a data file.

    Parameters
    ----------
    rf_name : resource_filename
        Name of resource file.

    Returns
    -------
    interp : interp1d
        An interpolator created using the first column of the file as the x
        values and second as the y values. interp will not raise a bounds error
        and uses a fill values of 0.0.
    """
    xs, ys = np.loadtxt(rf_name, delimiter=",").T
    return interp1d(xs, ys, bounds_error=bounds_error, fill_value=fill_value)


def spec_res_fn(ep, e, energy_res):
    """Get the spectral resolution function.
    """
    sigma = e * energy_res(e)

    if sigma == 0:
        if hasattr(ep, "__len__"):
            return np.zeros(ep.shape)
        else:
            return 0.0
    else:
        return (
            1.0
            / np.sqrt(2.0 * np.pi * sigma ** 2)
            * np.exp(-(ep - e) ** 2 / (2.0 * sigma ** 2))
        )


def convolved_spectrum_fn(
    e_min, e_max, e_cm, energy_res, spec_fn=None, line_fn=None, n_pts=1000
):
    r"""
    Convolves a continuum and line spectrum with a detector's spectral
    resolution function.

    Parameters
    ----------
    e_min : float
        Lower bound of energy range over which to perform convolution.
    e_max : float
        Upper bound of energy range over which to perform convolution.
    e_cm : float
        Center of mass energy for DM annihilation.
    energy_res : float -> float
        The detector's energy resolution (Delta E / E) as a function of
        photon energy in MeV.
    spec_fn : np.array -> np.array
        Continuum spectrum function.
    line_fn : np.array -> np.array
        Function returning a dict with information about spectral lines.
    n_pts : float
        Number of points to use to create resulting interpolating function.

    Returns
    -------
    dnde_conv : InterpolatedUnivariateSpline
        An interpolator giving the DM annihilation spectrum as seen by the
        detector. Using photon energies outside the range [e_min, e_max] will
        produce a ``bounds_errors``.
    """
    es = np.geomspace(e_min, e_max, n_pts)
    dnde_conv = np.zeros(es.shape)

    # Pad energy grid to avoid edge effects
    es_padded = np.geomspace(0.1 * e_min, 10 * e_max, n_pts)
    if spec_fn is not None:
        dnde_src = spec_fn(es_padded, e_cm)
        if not np.all(dnde_src == 0):

            def integral(e):
                """
                Performs the integration at given photon energy.
                """
                spec_res_fn_vals = spec_res_fn(es_padded, e, energy_res)
                integrand_vals = (
                    dnde_src * spec_res_fn_vals / trapz(spec_res_fn_vals, es_padded)
                )

                return trapz(integrand_vals, es_padded)

            dnde_conv += np.vectorize(integral)(es)

    # Line contribution
    if line_fn is not None:
        for ch, line in line_fn(e_cm).items():
            dnde_conv += (
                line["bf"]
                * spec_res_fn(es, line["energy"], energy_res)
                * (2.0 if ch == "g g" else 1.0)
            )

    return InterpolatedUnivariateSpline(es, dnde_conv, k=1, ext="raise")
